from django.contrib.gis.shortcuts import compress_kml
from django.template import loader, Context
from rest_framework.renderers import BaseRenderer

from spillway.collections import Feature, FeatureCollection


class BaseGeoRenderer(BaseRenderer):
    """Base renderer for geographic features."""

    def _features(self, data):
        if not isinstance(data, (Feature, FeatureCollection)):
            if isinstance(data, dict):
                data = Feature(**data)
            else:
                data = FeatureCollection(features=data)
        return data


class GeoJSONRenderer(BaseGeoRenderer):
    """Renderer which serializes to GeoJSON.

    This renderer purposefully avoids reserialization of GeoJSON from the
    spatial backend which greatly improves performance.
    """
    media_type = 'application/vnd.geo+json'
    format = 'geojson'

    def render(self, data, accepted_media_type=None, renderer_context=None):
        """Returns *data* encoded as GeoJSON."""
        return str(self._features(data))


class TemplateRenderer(BaseGeoRenderer):
    """Template based feature renderer."""
    template_name = None

    def render(self, data, accepted_media_type=None, renderer_context=None):
        collection = self._features(data)
        try:
            features = collection['features']
        except KeyError:
            features = [collection]
        template = loader.get_template(self.template_name)
        return template.render(Context({'features': features}))


class KMLRenderer(TemplateRenderer):
    """Renderer which serializes to KML."""
    media_type = 'application/vnd.google-earth.kml+xml'
    format = 'kml'
    template_name = 'spillway/placemarks.kml'


class KMZRenderer(KMLRenderer):
    """Renderer which serializes to KMZ."""
    media_type = 'application/vnd.google-earth.kmz'
    format = 'kmz'

    def render(self, *args, **kwargs):
        kmldata = super(KMZRenderer, self).render(*args, **kwargs)
        return compress_kml(kmldata)


class SVGRenderer(TemplateRenderer):
    """Renderer which serializes to SVG."""
    media_type = 'image/svg+xml'
    format = 'svg'
    template_name = 'spillway/features.svg'



import math

from django.shortcuts import get_object_or_404
from rest_framework.response import Response
from rest_framework.generics import GenericAPIView, ListAPIView

from spillway import forms, renderers
from spillway.generics import BaseGeoView


class MapView(GenericAPIView):
    """View for rendering map tiles from /{z}/{x}/{y}/ tile coordinates."""
    renderer_classes = (renderers.MapnikRenderer, renderers.MapnikJPEGRenderer)

    # FIXME: instead of here, validate form wherever self.kwargs is set, in
    # initial() is it?
    def get_renderer_context(self):
        context = super(MapView, self).get_renderer_context()
        form = forms.MapTile(dict(self.request.QUERY_PARAMS.dict(),
                                  **context.pop('kwargs')))
        context.update(form.cleaned_data if form.is_valid() else {})
        return context

    def get(self, request, *args, **kwargs):
        return Response(self.get_object())
        #print 'FORMAT:', self.format_kwarg
        obj = self.get_object()
        print 'OBJ:', type(obj), obj
        return Response(obj)
        # FIXME: inspect renderer context
        #args = self.kwargs.get('band') or ()
        #return Response(self.get_object().layer(*args))

    #def get_object(self, queryset=None):
    def _noget_object(self):
        queryset = self.filter_queryset(self.get_queryset())
        #filter = {}
        #for field in self.lookup_fields:
            #filter[field] = self.kwargs[field]
        #return get_object_or_404(queryset, **filter)
        #return get_object_or_404(queryset, **self.kwargs)
        #query = {k: v for k, v in self.kwargs.items() if k not in 'xyz'}
        querykeys = self.kwargs.viewkeys() - {'x', 'y', 'z', 'format'}
        print 'QUERY', querykeys
        #query = dict(zip(querykeys, map(self.kwargs.get, querykeys))
        query = {k: v for k, v in self.kwargs.items() if k in querykeys}
        return get_object_or_404(queryset, **query)


class TileView(BaseGeoView, ListAPIView):
    """View for serving tiled GeoJSON from a GeoModel."""
    paginate_by = None
    renderer_classes = (renderers.GeoJSONRenderer,)
    # Geometry simplification tolerances based on tile zlevel, see
    # http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames.
    #SpatialReference(3857).GetSemiMajor() == 6378137.0
    tolerances = [6378137 * 2 * math.pi / (2 ** (zoom + 8))
                  for zoom in range(20)]

    def filter_queryset(self, queryset):
        queryset = super(TileView, self).filter_queryset(queryset)
        form = forms.MapTile(self.kwargs)
        params = form.cleaned_data if form.is_valid() else {}
        bbox = params.get('bbox')
        # Tile grid uses 3857, but coordinates should be in 4326 commonly.
        tile_srid = 3857
        coord_srid = bbox.srid
        original_srid = queryset.geo_field.srid
        try:
            tolerance = self.tolerances[params['z']]
        except IndexError:
            tolerance = self.tolerances[-1]
        geom_wkt = bbox.ewkt
        queryset = (queryset.filter_geometry(intersects=geom_wkt)
                            .intersection(geom_wkt))
        for obj in queryset:
            geom = obj.intersection
            # Geometry must be in Web Mercator for simplification.
            if geom.srid != tile_srid:
                # Result of intersection does not have SRID set properly.
                if geom.srid is None:
                    geom.srid = original_srid
                geom.transform(tile_srid)
            geom = geom.simplify(tolerance, preserve_topology=True)
            geom.transform(coord_srid)
            obj.geojson = geom.geojson
        return queryset



from rest_framework.generics import ListAPIView, ListCreateAPIView, RetrieveAPIView
from rest_framework.settings import api_settings

from spillway import filters, forms, mixins, renderers, serializers

_default_renderers = tuple(api_settings.DEFAULT_RENDERER_CLASSES)


class BaseGeoView(mixins.QueryFormMixin):
    """Base view for models with geometry fields."""
    model_serializer_class = serializers.FeatureSerializer
    pagination_serializer_class = serializers.PaginatedFeatureSerializer
    query_form_class = forms.GeometryQueryForm
    renderer_classes = _default_renderers + (
        renderers.GeoJSONRenderer, renderers.KMLRenderer, renderers.KMZRenderer)

    #def get_serializer_class(self):
        #serializer_class = super(BaseGeoView, self).get_serializer_class()
        ## Use typical record-based serialization for default renderers,
        ## otherwise use feature-based for geoformats.
        #if self.wants_default_renderer():
            #serializer_class = serializers.GeoModelSerializer
            #serializer_class.Meta.model = self.model
        #return serializer_class

    def wants_default_renderer(self):
        """Returns true when using a default renderer class."""
        return isinstance(self.request.accepted_renderer, _default_renderers)


class GeoDetailView(BaseGeoView, RetrieveAPIView):
    """Generic detail view providing vector geometry representations."""


# FIXME: All generics should follow drf naming, Geo(m)ListAPIView
class GeoListView(BaseGeoView, ListAPIView):
    """Generic list view providing vector geometry representations."""
    filter_backends = (filters.SpatialLookupFilter, filters.GeoQuerySetFilter)


class GeoListCreateAPIView(BaseGeoView, ListCreateAPIView):
    filter_backends = (filters.SpatialLookupFilter, filters.GeoQuerySetFilter)


class BaseRasterView(BaseGeoView):
    """Base view for raster models."""
    model_serializer_class = serializers.RasterModelSerializer
    query_form_class = forms.RasterQueryForm

    def get_serializer_context(self):
        context = super(BaseRasterView, self).get_serializer_context()
        context.update(params=self.clean_params())
        return context


class RasterDetailView(BaseRasterView, RetrieveAPIView):
    """View providing access to a Raster model instance."""
    renderer_classes = _default_renderers + (
        renderers.GeoTIFFRenderer,
        renderers.HFARenderer,
    )


class RasterListView(BaseRasterView, ListAPIView):
    """View providing access to a Raster model QuerySet."""
    filter_backends = (filters.SpatialLookupFilter,)
    renderer_classes = _default_renderers + (
        renderers.GeoTIFFZipRenderer,
        renderers.HFAZipRenderer,
    )



from django.contrib.gis.db import models
from rest_framework import serializers, pagination
from greenwich.srs import SpatialReference

from spillway.collections import Feature, FeatureCollection, NamedCRS
from spillway.fields import GeometryField, NDArrayField


class GeoModelSerializerOptions(serializers.ModelSerializerOptions):
    def __init__(self, meta):
        super(GeoModelSerializerOptions, self).__init__(meta)
        self.geom_field = getattr(meta, 'geom_field', None)


class GeoModelSerializer(serializers.ModelSerializer):
    """Serializer class for GeoModels."""
    _options_class = GeoModelSerializerOptions
    field_mapping = dict({
        models.GeometryField: GeometryField,
        models.PointField: GeometryField,
        models.LineStringField: GeometryField,
        models.PolygonField: GeometryField,
        models.MultiPointField: GeometryField,
        models.MultiLineStringField: GeometryField,
        models.MultiPolygonField: GeometryField,
        models.GeometryCollectionField: GeometryField
    }, **serializers.ModelSerializer.field_mapping)

    def get_default_fields(self):
        """Returns a fields dict for this serializer with a 'geometry' field
        added.
        """
        fields = super(GeoModelSerializer, self).get_default_fields()
        # Set the geometry field name when it's undeclared.
        if not self.opts.geom_field:
            for name, field in fields.items():
                if isinstance(field, GeometryField):
                    self.opts.geom_field = name
        return fields

class CRSField(serializers.Field):
    crscount = 0
    def to_native(self, value):
        #import ipdb; ipdb.set_trace()
        self.crscount += 1
        print 'COUNT:', self.crscount
        #return value
        return 4326

class FeatureSerializer(GeoModelSerializer):
    #crs = CRSField(source='*')
    def __init__(self, *args, **kwargs):
        super(FeatureSerializer, self).__init__(*args, **kwargs)
        self.fields[self.opts.geom_field].set_default_source()
        #try:
            ##self._crs = self.init_data['crs']
            #self.init_data = self.init_data['features']
        #except KeyError:
            #pass
        #view = self.context.get('view')
        #try:
            #wants_default_renderer = view.wants_default_renderer()
        #except AttributeError:
            #wants_default_renderer = True
        ## Alter the field source based on geometry output format.
        #if not wants_default_renderer:
            #renderer = view.request.accepted_renderer
            #self.fields[self.opts.geom_field].set_source(renderer.format)

    @property
    def data(self):
        if self._data is None:
            data = super(FeatureSerializer, self).data
            fieldname = self.opts.geom_field
            if self.many or isinstance(data, (list, tuple)):
                try:
                    srid = (self.object.query.transformed_srid or
                            self.object.geo_field.srid)
                except AttributeError:
                    srid = None
                self._data = FeatureCollection(features=data, crs=srid)
            else:
                try:
                    geom = getattr(self.object, fieldname)
                except AttributeError:
                    pass
                else:
                    self._data['crs'] = NamedCRS(geom.srid)
        return self._data

    #@property
    #def errors(self):

    def to_native(self, obj):
        native = super(FeatureSerializer, self).to_native(obj)
        geometry = native.pop(self.opts.geom_field)
        pk = native.pop(obj._meta.pk.name, None)
        return Feature(pk, geometry, native)

    # FIXME: How to handle many?
    def from_native(self, data, files=None):
        if data and 'features' in data:
            for feat in data['features']:
                return self.from_native(feat, files)
        try:
            sref = SpatialReference(data['crs']['properties']['name'])
        except KeyError:
            sref = None
        record = {self.opts.geom_field: data.get('geometry')}
        record.update(data.get('properties', {}))
        #FIXME: handle reprojection.
        feature = super(FeatureSerializer, self).from_native(record, files)
        if feature and sref:
            geom = getattr(feature, self.opts.geom_field)
            geom.srid = sref.srid
        return feature

class FeatureField(serializers.Field):
    def to_native(self, obj):
        native = super(FeatureField, self).to_native(obj)
        geometry = native.pop(self.opts.geom_field)
        pk = native.pop(obj._meta.pk.name, None)
        return Feature(pk, geometry, native)
class NFeatureSerializer(serializers.Serializer):
    def to_native(self, obj):
        native = super(NFeatureSerializer, self).to_native(obj)
        #geometry = native.pop(self.opts.geom_field)
        import ipdb; ipdb.set_trace()
        pk = native.pop(obj._meta.pk.name, None)
        return Feature(pk, geometry, native)

#FeatureCollectionSerializer({'features': GeoModel.objects.all()})
class FeatureCollectionSerializer(serializers.Serializer):
    #features = FeatureSerializer(many=True) # needs model set
    features = NFeatureSerializer(source='*')
    #features = FeatureField(source='*') # nested list
    #bbox = serializers.CharField()
    crs = CRSField(source='*')

    #def to_native(self, obj):
        #native = super(FeatureCollectionSerializer, self).to_native(obj)
        #import ipdb; ipdb.set_trace()
        #return native

class PaginatedFeatureSerializer(pagination.PaginationSerializer):
    results_field = 'features'


class RasterModelSerializerOptions(GeoModelSerializerOptions):
    def __init__(self, meta):
        super(RasterModelSerializerOptions, self).__init__(meta)
        self.raster_field = getattr(meta, 'raster_field', None)


class RasterModelSerializer(GeoModelSerializer):
    _options_class = RasterModelSerializerOptions

    #def __init__(self, *args, **kwargs):
        #rformat = kwargs.pop('format', None)
        #super(RasterModelSerializer, self).__init__(*args, **kwargs)
        #if not rformat:
            #request = self.context.get('request')
            #if request:
                #rformat = request.accepted_renderer.format
        #self.format = rformat
        ##self.fields['image'].format = rformat

    def get_default_fields(self):
        fields = super(RasterModelSerializer, self).get_default_fields()
        if not self.opts.raster_field:
            for name, field in fields.items():
                if isinstance(field, serializers.FileField):
                    self.opts.raster_field = name
        request = self.context.get('request')
        render_format = request.accepted_renderer.format if request else None
        # Serialize image data as arrays when json is requested.
        if render_format == 'json':
            fields[self.opts.raster_field] = NDArrayField()
        elif render_format in ('api', 'html'):
            pass
        elif self.opts.raster_field and 'path' not in fields:
            # Add a filepath field for GDAL based renderers.
            fields['path'] = serializers.CharField(
                source='%s.path' % self.opts.raster_field)
        return fields

    #def to_native(self, obj):
        #params = self.context.get('params', {})
        #geom = params.get('g')
        ##request = self.context.get('request')
        ##format = request and request.accepted_renderer.format
        #if callable(getattr(obj, 'clip', None)):
            #clip = obj.clip(geom, self.format)
            ##import ipdb; ipdb.set_trace()
            #if clip:
                #setattr(obj, self.opts.raster_field, clip)
        #return super(RasterModelSerializer, self).to_native(obj)


class RasterCatalogSerializer(serializers.HyperlinkedModelSerializer):
    pass


#class RasterStoreSerializer(serializers.HyperlinkedModelSerializer):
    ##tile_url = serializers.URLField(read_only=True)
    ##source = serializers.URLField(source='gdal.url', read_only=True)
    #path = serializers.CharField(source='gdal.path', read_only=True)
    ##source = serializers.ImageField()
    ##source = serializers.FileField()

    #class Meta:
        ##model = RasterStore
        #exclude = ('gdal', 'proj4')


#{'map': 'map.xml', 'style': 'default', 'layer': obj.layer, 'srs': 'epsg:3857'}
class MapnikSerializer(serializers.Serializer):
    mapfile = serializers.CharField(required=False)
    bbox = GeometryField(required=False)
    x = serializers.IntegerField()
    y = serializers.IntegerField()
    z = serializers.IntegerField()
    #band = serializers.IntegerField(default=1)
    style = serializers.CharField(required=False)
    #srs =

    def to_native(self, obj):
        obj = super(MapnikSerializer, self).to_native(obj)
        nikmap = mapnik.Map(256, 256)
        nikmap.buffer_size = 128
        mapnik.load_map(nikmap, mapfile)
        layer = obj.layer()
        x, y, z = map(obj.get, ('x', 'y', 'z'))
        obj['bbox'] = transform_tile(x, y, z) + transform_tile(x + 1, y + 1, z)
        style = nikmap.find_style('default')
        rule = style.rules[0]
        symbolizer = rule.symbols[0]
        # Generate equal interval bins from min/max pixel values.
        breaks = np.linspace(self.object.minval, self.object.maxval,
                             len(symbolizer.colorizer.stops))
        # Update the stop values for the color ramp.
        for stop, breakval in zip(symbolizer.colorizer.stops, breaks):
            stop.value = breakval
        #layer.styles.append(style.name)
        layer.styles.append('default')
        # Must append layer to map *after* appending style to it.
        nikmap.layers.append(layer)
        bbox.transform(nikmap.srs)
        nikmap.zoom_to_box(mapnik.Box2d(*bbox.extent))
        img = mapnik.Image(nikmap.width, nikmap.height)
        mapnik.render(nikmap, img)
        return img.tostring('png')

    #def from_native(self, obj, files=None):
        #print 'OBJECT', obj
    #def restore_object(self, attrs, instance=None):
        #nikmap = mapnik.Map(256, 256)
        #nikmap.buffer_size = 128
        #mapnik.load_map(nikmap, mapfile)
        #return nikmap



from spillway.compat import mapnik

# ColorBrewer http://colorbrewer2.org/
colors = {'YlGnBu': ('#ffffd9', '#edf8b1', '#c7e9b4', '#7fcdbb', '#41b6c4',
                     '#1d91c0', '#225ea8', '#253494', '#081d58'),
          'YlGn': ('#ffffe5', '#f7fcb9', '#d9f0a3', '#addd8e', '#78c679',
                   '#41ab5d', '#238443', '#006837', '#004529'),
          'YlGnBu': ('#ffffd9', '#edf8b1', '#c7e9b4', '#7fcdbb', '#41b6c4',
                     '#1d91c0', '#225ea8', '#253494', '#081d58'),
          'GnBu': ('#f7fcf0', '#e0f3db', '#ccebc5', '#a8ddb5', '#7bccc4',
                   '#4eb3d3', '#2b8cbe', '#0868ac', '#084081'),
          'BuGn': ('#f7fcfd', '#e5f5f9', '#ccece6', '#99d8c9', '#66c2a4',
                   '#41ae76', '#238b45', '#006d2c', '#00441b'),
          'PuBuGn': ('#fff7fb', '#ece2f0', '#d0d1e6', '#a6bddb', '#67a9cf',
                     '#3690c0', '#02818a', '#016c59', '#014636'),
          'PuBu': ('#fff7fb', '#ece7f2', '#d0d1e6', '#a6bddb', '#74a9cf',
                   '#3690c0', '#0570b0', '#045a8d', '#023858'),
          'BuPu': ('#f7fcfd', '#e0ecf4', '#bfd3e6', '#9ebcda', '#8c96c6',
                   '#8c6bb1', '#88419d', '#810f7c', '#4d004b'),
          'RdPu': ('#fff7f3', '#fde0dd', '#fcc5c0', '#fa9fb5', '#f768a1',
                   '#dd3497', '#ae017e', '#7a0177', '#49006a'),
          'PuRd': ('#f7f4f9', '#e7e1ef', '#d4b9da', '#c994c7', '#df65b0',
                   '#e7298a', '#ce1256', '#980043', '#67001f'),
          'OrRd': ('#fff7ec', '#fee8c8', '#fdd49e', '#fdbb84', '#fc8d59',
                   '#ef6548', '#d7301f', '#b30000', '#7f0000'),
          'YlOrRd': ('#ffffcc', '#ffeda0', '#fed976', '#feb24c', '#fd8d3c',
                     '#fc4e2a', '#e31a1c', '#bd0026', '#800026'),
          'YlOrBr': ('#ffffe5', '#fff7bc', '#fee391', '#fec44f', '#fe9929',
                     '#ec7014', '#cc4c02', '#993404', '#662506'),
          'Purples': ('#fcfbfd', '#efedf5', '#dadaeb', '#bcbddc', '#9e9ac8',
                      '#807dba', '#6a51a3', '#54278f', '#3f007d'),
          'Blues': ('#f7fbff', '#deebf7', '#c6dbef', '#9ecae1', '#6baed6',
                    '#4292c6', '#2171b5', '#08519c', '#08306b'),
          'Greens': ('#f7fcf5', '#e5f5e0', '#c7e9c0', '#a1d99b', '#74c476',
                     '#41ab5d', '#238b45', '#006d2c', '#00441b'),
          'Oranges': ('#fff5eb', '#fee6ce', '#fdd0a2', '#fdae6b', '#fd8d3c',
                      '#f16913', '#d94801', '#a63603', '#7f2704'),
          'Reds': ('#fff5f0', '#fee0d2', '#fcbba1', '#fc9272', '#fb6a4a',
                   '#ef3b2c', '#cb181d', '#a50f15', '#67000d'),
          'Greys': ('#ffffff', '#f0f0f0', '#d9d9d9', '#bdbdbd', '#969696',
                    '#737373', '#525252', '#252525', '#000000'),
          'PuOr': ('#7f3b08', '#b35806', '#e08214', '#fdb863', '#fee0b6',
                   '#f7f7f7', '#d8daeb', '#b2abd2', '#8073ac', '#542788',
                   '#2d004b'),
          'BrBG': ('#543005', '#8c510a', '#bf812d', '#dfc27d', '#f6e8c3',
                   '#f5f5f5', '#c7eae5', '#80cdc1', '#35978f', '#01665e',
                   '#003c30'),
          'PRGn': ('#40004b', '#762a83', '#9970ab', '#c2a5cf', '#e7d4e8',
                   '#f7f7f7', '#d9f0d3', '#a6dba0', '#5aae61', '#1b7837',
                   '#00441b'),
          'PiYG': ('#8e0152', '#c51b7d', '#de77ae', '#f1b6da', '#fde0ef',
                   '#f7f7f7', '#e6f5d0', '#b8e186', '#7fbc41', '#4d9221',
                   '#276419'),
          'RdBu': ('#67001f', '#b2182b', '#d6604d', '#f4a582', '#fddbc7',
                   '#f7f7f7', '#d1e5f0', '#92c5de', '#4393c3', '#2166ac',
                   '#053061'),
          'RdGy': ('#67001f', '#b2182b', '#d6604d', '#f4a582', '#fddbc7',
                   '#ffffff', '#e0e0e0', '#bababa', '#878787', '#4d4d4d',
                   '#1a1a1a'),
          'RdYlBu': ('#a50026', '#d73027', '#f46d43', '#fdae61', '#fee090',
                     '#ffffbf', '#e0f3f8', '#abd9e9', '#74add1', '#4575b4',
                     '#313695'),
          'Spectral': ('#9e0142', '#d53e4f', '#f46d43', '#fdae61', '#fee08b',
                       '#ffffbf', '#e6f598', '#abdda4', '#66c2a5', '#3288bd',
                       '#5e4fa2'),
          'RdYlGn': ('#a50026', '#d73027', '#f46d43', '#fdae61', '#fee08b',
                     '#ffffbf', '#d9ef8b', '#a6d96a', '#66bd63', '#1a9850',
                     '#006837'),
          'Accent': ('#7fc97f', '#beaed4', '#fdc086', '#ffff99', '#386cb0',
                     '#f0027f', '#bf5b17', '#666666'),
          'Dark2': ('#1b9e77', '#d95f02', '#7570b3', '#e7298a', '#66a61e', '#e6ab02',
                    '#a6761d', '#666666'),
          'Paired': ('#a6cee3', '#1f78b4', '#b2df8a', '#33a02c', '#fb9a99', '#e31a1c',
                     '#fdbf6f', '#ff7f00', '#cab2d6', '#6a3d9a', '#ffff99', '#b15928'),
          'Pastel1': ('#fbb4ae', '#b3cde3', '#ccebc5', '#decbe4', '#fed9a6', '#ffffcc',
                      '#e5d8bd', '#fddaec', '#f2f2f2'),
          'Pastel2': ('#b3e2cd', '#fdcdac', '#cbd5e8', '#f4cae4', '#e6f5c9', '#fff2ae',
                      '#f1e2cc', '#cccccc'),
          'Set1': ('#e41a1c', '#377eb8', '#4daf4a', '#984ea3', '#ff7f00', '#ffff33',
                   '#a65628', '#f781bf', '#999999'),
          'Set2': ('#66c2a5', '#fc8d62', '#8da0cb', '#e78ac3', '#a6d854', '#ffd92f',
                   '#e5c494', '#b3b3b3'),
          'Set3': ('#8dd3c7', '#ffffb3', '#bebada', '#fb8072', '#80b1d3', '#fdb462',
                   '#b3de69', '#fccde5', '#d9d9d9', '#bc80bd', '#ccebc5', '#ffed6f'),
          # Generic Mapping Tools (GMT) http://gmt.soest.hawaii.edu/
          # FIXME: Start using css for these lengthy gradients, call static template tag
          # to get css file path.
          'Haxby': ('#0a0079', '#280096', '#1405af', '#000ac8', '#0019d4', '#0028e0', '#1a66f0',
'#0d81f8', '#19afff', '#32beff', '#44caff', '#61e1f0', '#6aebe1', '#7cebc8',
'#8aecae', '#acf5a8', '#cdffa2', '#dff58d', '#f0ec79', '#f7d768', '#ffbd57',
'#ffa045', '#f4754b', '#ee504e', '#ff5a5a', '#ff7c7c', '#ff9e9e', '#f5b3ae',
'#ffc4c4', '#ffd7d7', '#ffebeb', '#fffefd')}
# Add reversed colors.
colors.update({'%s_r' % k: tuple(reversed(v)) for k, v in colors.items()})

def add_colorizer_stops(style, bins, mcolors):
    rule = style.rules[0]
    symbolizer = rule.symbols[0]
    for value, color in zip(bins, mcolors):
        symbolizer.colorizer.add_stop(value, mapnik.Color(color))
    return style

def make_raster_style():
    """Returns a default raster Style."""
    style = mapnik.Style()
    rule = mapnik.Rule()
    symbolizer = mapnik.RasterSymbolizer()
    symbolizer.colorizer = mapnik.RasterColorizer(
        mapnik.COLORIZER_LINEAR, mapnik.Color(0, 0, 0, 255))
    rule.symbols.append(symbolizer)
    style.rules.append(rule)
    return style

#class RasterLayer(mapnik.Layer):
    #def layer(self, band=1):
        #layer = mapnik.Layer(str(self), SpatialReference(self.srs).proj4)
        #layer.datasource = mapnik.Gdal(file=self.image.path, band=band)
        #return layer

def raster_to_layer(raster, band=1):
    layer = mapnik.Layer(str(raster), SpatialReference(raster.srs).proj4)
    layer.datasource = mapnik.Gdal(file=raster.image.path, band=band)
    return layer

#def add_layer()
def _to_layer(object, stylename, canvas):
    try:
        style = canvas.find_style(stylename)
    except KeyError:
        style = make_raster_style()
        canvas.append_style(stylename, style)
        add_colorizer_stops(style, (object.minval, object.maxval),
                            name=stylename)
    try:
        layer = object.layer()
    except AttributeError:
        pass
    else:
        #if hasattr(style,
        #if layer.datasource.type() == mapnik.DataType.Raster:
            #styles.add_colorizer_stops(
                #style, (object.minval, object.maxval), name=stylename)
        #else: Vector?
        layer.styles.append(stylename)
        # Must append layer to map *after* appending style to it.
        canvas.layers.append(layer)

#def adjust_breaks(colorizer, extrema):
#def rebin()
#def adjust_bins(colorizer, extrema, nbins=None, binwidth=None):
#def adjust_to_minmax(colorizer, extrema, equal_width=True):
    ## Generate equal interval bins from min/max pixel values.
    #if equal_width:
        ##breaks = np.linspace(*extrema + (len(colorizer.stops),))
        ## Update the stop values for the color ramp.
        #for stop, breakval in zip(colorizer.stops, breaks):
            #stop.value = breakval
    #else:
        #colorizer.stops[0] =
    #return colorizer

#def adjust_intervals(colorizer, extrema):
#def adjust_to_minmax(colorizer, extrema):
#def adjust_equint(colorizer, extrema):
def equalize_intervals(style, values):
    """Adjust RasterColorizer to equal intervals."""
    rule = style.rules[0]
    symbolizer = rule.symbols[0]
    stops = symbolizer.colorizer.stops
    # Generate equal interval bins from min/max pixel values.
    breaks = np.linspace(values[0], values[-1], len(stops))
    # Update the stop values for the color ramp.
    for stop, breakval in zip(stops, breaks):
        stop.value = breakval
        print stop, breakval
    #return colorizer
    return style



from django.core.exceptions import ImproperlyConfigured

try:
    import simplejson as json
except ImportError:
    import json

try:
    import mapnik
except ImportError:
    class Mapnik(object):
        def __getattr__(self, attr):
            raise ImproperlyConfigured('Mapnik must be installed')
    mapnik = Mapnik()



import os
import tempfile
from wsgiref.util import FileWrapper
import zipfile

from greenwich.geometry import Geometry
from greenwich.io import MemFileIO
from greenwich.raster import Raster, driver_for_path
from rest_framework.renderers import BaseRenderer


class BaseGDALRenderer(BaseRenderer):
    """Abstract renderer which encodes to a GDAL supported raster format."""
    media_type = 'application/octet-stream'
    format = None
    charset = None
    render_style = 'binary'

    @property
    def file_ext(self):
        return os.extsep + os.path.splitext(self.format)[0]

    def basename(self, item):
        """Returns the output filename.

        Arguments:
        item -- dict containing 'path'
        """
        fname = os.path.basename(item['path'])
        return os.path.splitext(fname)[0] + self.file_ext

    def render(self, data, accepted_media_type=None, renderer_context=None):
        if isinstance(data, dict):
            data = [data]
        self.set_filename(self.basename(data[0]), renderer_context)
        img = self._render_items(data, renderer_context)[0]
        # File contents could contain null bytes but not file names.
        try:
            isfile = os.path.isfile(img)
        except TypeError:
            isfile = False
        if isfile:
            self.set_response_length(os.path.getsize(img), renderer_context)
            img = open(img)
        return FileWrapper(img)

    def _render_items(self, items, renderer_context):
        renderer_context = renderer_context or {}
        view = renderer_context.get('view')
        geom = view and view.clean_params().get('g')
        driver = driver_for_path(self.file_ext.replace(os.extsep, ''))
        if geom:
            # Convert to wkb for ogr.Geometry
            geom = Geometry(wkb=bytes(geom.wkb), srs=geom.srs.wkt)
        imgdata = []
        for item in items:
            imgpath = item['path']
            # No conversion is needed if the original format without clipping
            # is requested.
            if not geom and imgpath.endswith(self.file_ext):
                imgdata.append(imgpath)
                continue
            memio = MemFileIO()
            if geom:
                with Raster(imgpath) as r:
                    with r.clip(geom) as clipped:
                        clipped.save(memio, driver)
            else:
                driver.copy(imgpath, memio.name)
            imgdata.append(memio.read())
            memio.close()
        return imgdata

    def set_filename(self, name, renderer_context):
        type_name = 'attachment; filename=%s.%s' % (name, self.format)
        try:
            renderer_context['response']['Content-Disposition'] = type_name
        except (KeyError, TypeError):
            pass

    def set_response_length(self, length, renderer_context):
        try:
            renderer_context['response']['Content-Length'] = length
        except (KeyError, TypeError):
            pass


class HFARenderer(BaseGDALRenderer):
    """Renders a raster to Erdas Imagine (.img) format."""
    format = 'img'


class GeoTIFFRenderer(BaseGDALRenderer):
    """Renders a raster to GeoTIFF (.tif) format."""
    media_type = 'image/tiff'
    format = 'tif'


class GeoTIFFZipRenderer(BaseGDALRenderer):
    """Bundles GeoTIFF rasters in a zip archive."""
    media_type = 'application/zip'
    format = 'tif.zip'
    arcdirname = 'data'

    def render(self, data, accepted_media_type=None, renderer_context=None):
        if isinstance(data, dict):
            data = [data]
        rendered = self._render_items(data, renderer_context)
        self.set_filename(self.arcdirname, renderer_context)
        fp = tempfile.TemporaryFile(suffix=os.extsep + self.format)
        with zipfile.ZipFile(fp, mode='w') as zf:
            for raster, attrs in zip(rendered, data):
                arcname = os.path.join(self.arcdirname, self.basename(attrs))
                # Attempt to write from the filename first, or fall back to the
                # file contents.
                try:
                    zf.write(raster, arcname=arcname)
                except TypeError:
                    zf.writestr(arcname, raster)
        self.set_response_length(fp.tell(), renderer_context)
        fp.seek(0)
        return FileWrapper(fp)


class HFAZipRenderer(GeoTIFFZipRenderer):
    """Bundles Erdas Imagine rasters in a zip archive."""
    format = 'img.zip'



"""Serializer fields"""
from django.contrib.gis import forms
from rest_framework.fields import FileField, WritableField
from greenwich import io, Raster

from spillway.compat import json


class GeometryField(WritableField):
    type_name = 'GeometryField'
    type_label = 'geometry'
    form_field_class = forms.GeometryField
    # TODO: Introspect available formats from spatial backends.
    _formats = ('geohash', 'geojson', 'gml', 'kml', 'svg')
    #_formats = ('geohash', 'geojson', 'gml', 'json', 'kml', 'svg')

    def set_default_source(self):
        """Set a default source based on the selected renderer."""
        # Do not override an explicitly provided field source.
        if self.source:
            return
        view = self.context.get('view')
        # Alter the field source based on geometry output format.
        try:
            wants_default_renderer = view.wants_default_renderer()
        except AttributeError:
            wants_default_renderer = True
        # Alter the field source based on geometry output format.
        if not wants_default_renderer:
            renderer = view.request.accepted_renderer
            if renderer.format in self._formats:
                self.source = renderer.format
                # Single objects must use GEOSGeometry attrs to provide formats.
                if not self.parent.many:
                    self.source = '%s.%s' % (self.label, self.source)

    #def set_format(self, format):
    #def set_source(self, format):
    def use_format(self, format):
        if self.source:
            return
        if format in self._formats:
            self.source = format
            # Single objects must use GEOSGeometry attrs to provide formats.
            if not self.parent.many:
                self.source = '%s.%s' % (self.label, self.source)

    def to_native(self, value):
        # Create a dict from the GEOSGeometry when the value is not previously
        # serialized from the spatial db.
        try:
            return {'type': value.geom_type, 'coordinates': value.coords}
        # Value is already serialized as geojson, kml, etc.
        except AttributeError:
            return value

    def from_native(self, value):
        # forms.GeometryField cannot handle geojson dicts.
        if isinstance(value, dict):
            value = json.dumps(value)
        return super(GeometryField, self).from_native(value)


class NDArrayField(FileField):
    type_name = 'NDArrayField'
    type_label = 'ndarray'

    def to_native(self, value):
        params = self.context.get('params', {})
        geom = params.get('g')
        #band = params.get('band')
        with Raster(getattr(value, 'path', value)) as r:
            #if band:
                #rband = r[band]
            #arr = r.clip(geom).masked_array() if geom else r.array()
            arr = r.clip(geom).array() if geom else r.array()
            if not geom:
                return r.array().tolist()
            with r.clip(geom) as clipped:
                return clipped.array().tolist()
        #r = Raster(getattr(value, 'path', value))
        #arr = r.clip(geom).masked_array() if geom else r.array()
        #arr = r.clip(geom).array() if geom else r.array()
        #r.close()
        #func = params.get('filter_func')
        # FIXME: Handle ints, float32 has rounding issues, why do we have to
        # cast to float64?
        #return (func(arr) if func else arr).astype(float).round(4).tolist()
        #return arr.tolist()


class GDALField(FileField):
    type_name = 'GDALField'
    type_label = 'gdaldataset'

    #def initialize(self, *args, **kwargs):
        #super(GDALField, self).initialize(*args, **kwargs)

    #def to_native(self, value):
        #return value.path

    def not_to_native(self, value):
        print 'VALUE', type(value), value
        #if value is None: #FIXME: field_to_native should not call this with None?
            #raise Exception
        params = self.context.get('params', {})
        geom = params.get('g')
        request = self.context.get('request')
        format = request and request.accepted_renderer.format or getattr(self, 'format', None)
        #if not request:
        if not format:
            return value.path
        #format = request.accepted_renderer.format
        assert '.' in format
        # No conversion is needed if the original format without clipping
        # is requested.
        if not geom and value.name.endswith(format):
            return value.path
        memio = io.MemFileIO(suffix=format)
        with Raster(value.path) as r:
            if geom:
                with r.clip(geom) as clipped:
                    clipped.save(memio)
            else:
                r.save(memio)
        imgdata = memio.read()
        memio.close()
        return imgdata



from django.contrib.gis import geos
from django.contrib.gis.db.models import query
from django.db import connection

# Many GeoQuerySet methods cannot be chained as expected and extending
# GeoQuerySet to work properly with serialization calls like
# .simplify(100).svg() will require patching
# django.contrib.gis.db.models.query. Work around this with a custom
# GeoQuerySet for now.
class GeoQuerySet(query.GeoQuerySet):
    """Extends the default GeoQuerySet with some unimplemented PostGIS
    functionality.
    """
    _scale = '%s(%%s, %%s, %%s)' % connection.ops.scale
    # Geometry outputs
    _formats = {'geojson': '%s(%%s, %%s)' % connection.ops.geojson,
                #'json': '%s(%%s, %%s)' % connection.ops.geojson,
                'kml': '%s(%%s, %%s)' % connection.ops.kml,
                'svg': '%s(%%s, 0, %%s)' % connection.ops.svg}
    # Json is an alias for geojson.
    #_formats['json'] = _formats['geojson']

    def _as_format(self, sql, format=None, precision=6):
        val = self._formats.get(format, self._formats['geojson'])
        return self.extra(select={format: val % (sql, precision)})

    def _transform(self, colname, srid=None):
        # FIXME: Used for different purposes?!
        #self.query.transformed_srid = srid
        # Transform for the side effect of changing geom srid.
        if srid:
            self.transform(srid)
        return ('%s(%s, %s)' % (connection.ops.transform, colname, srid)
                if srid else colname)

    def _simplify(self, colname, tolerance=0.0):
        # connection.ops does not have simplify available for PostGIS.
        return ('ST_Simplify(%s, %s)' % (colname, tolerance)
                if tolerance else colname)

    def extent(self, srid=None):
        """Returns the GeoQuerySet extent as a 4-tuple.

        The method chaining approach of
        geoqset.objects.transform(srid).extent() returns the extent in the
        original coordinate system, this method allows for transformation.

        Keyword args:
        srid -- EPSG id for for transforming the output geometry.
        """
        if not srid and not connection.ops.spatialite:
            return super(GeoQuerySet, self).extent()
        transform = self._transform(self.geo_field.column, srid)
        # Spatialite extent() is supported post-1.7.
        if connection.ops.spatialite:
            ext = {'extent': 'AsText(%s(%s))' % ('Extent', transform)}
        else:
            ext = {'extent': '%s(%s)' % (connection.ops.extent, transform)}
        try:
            # The bare order_by() is needed to remove the default sort field
            # which is not present in this aggregation.
            extent = (self.extra(select=ext)
                          .values_list('extent', flat=True)
                          .order_by()[0])
        except IndexError:
            return ()
        try:
            return connection.ops.convert_extent(extent)
        except NotImplementedError:
            return geos.GEOSGeometry(extent, srid).extent

    def filter_geometry(self, **kwargs):
        """Convenience method for providing spatial lookup types as keywords
        without underscores instead of the usual "geometryfield__lookuptype"
        format.
        """
        fieldname = self.geo_field.name
        query = {'%s__%s' % (fieldname, key): val
                 for key, val in kwargs.items()}
        return self.filter(**query)

    @property
    def geo_field(self):
        return self.query._geo_field()

    def scale(self, x, y, z=0.0, tolerance=0.0, precision=6, srid=None,
              format=None, **kwargs):
        """Returns a GeoQuerySet with scaled and optionally reprojected and
        simplified geometries, serialized to a supported format.
        """
        if not any((tolerance, srid, format)):
            return super(GeoQuerySet, self).scale(x, y, z, **kwargs)
        transform = self._transform(self.geo_field.column, srid)
        scale = self._scale % (transform, x, y)
        simplify = self._simplify(scale, tolerance)
        return self._as_format(simplify, format, precision)

    def simplify(self, tolerance=0.0, srid=None, format=None, precision=6):
        """Returns a GeoQuerySet with simplified geometries serialized to
        a supported geometry format.
        """
        # Transform first, then simplify.
        transform = self._transform(self.geo_field.column, srid)
        if format:
            simplify = (self._simplify(transform, tolerance)
                        if tolerance else transform)
            return self._as_format(simplify, format, precision)
        simplify = self._simplify(transform, tolerance)
        #FIXME: Should be "name" not "column" right?
        # TODO: EWKB bug is fixed in spatialite 4.2+ so this can be removed.
        if connection.ops.spatialite:
            simplify = 'AsEWKT(%s)' % simplify
        return self.extra(select={self.geo_field.column: simplify})


        # Simplify, then transform.
        simplify = self._simplify(self.geo_field.column, tolerance)
        transform = self._transform(simplify, srid) if srid else simplify
        if format:
            return self._as_format(transform, format, precision)
        #FIXME: Should be "name" not "column" right?
        return self.extra(select={self.geo_field.column: self._wkb(transform)})

    # FIXME: add preserve_topology
    def nosimplify(self, tolerance=0.0, srid=None, format=None, precision=6):
        #'self.query.extra': {'intersection': (u'ST_Intersection("coredata_ecoregion"."the_geom",%s)',
        #import ipdb; ipdb.set_trace()
        try:
            attr, query = self.query.extra.popitem()
        except KeyError:
            #simplify = self._simplify(self.geo_field.column, tolerance)
            attr = self.geo_field.name
            query = (self.geo_field.column, '') # need the postgis adapter object.

        simplify = self._simplify(query[0], tolerance)
        transform = self._transform(simplify, srid) if srid else simplify
        if format:
            val = self._formats.get(format, self._formats['geojson'])
            geoformat = val % (transform, precision)
            self.query.extra[format] = (geoformat, query[1])
            #return self
            return self._clone()
        #return self.extra(select={self.geo_field.column: self._wkb(transform)})
        self.query.extra[attr] = (self._wkb(transform), query[1])
        #return self
        return self._clone()

    #def _merge_extra(self, vals):
        #self.query.extra

    #def tile(self, tilecoord):

    def _no_transform(self, srid=4326, **kwargs):
        """
        Transforms the given geometry field to the given SRID.  If no SRID is
        provided, the transformation will default to using 4326 (WGS84).
        """
        import ipdb; ipdb.set_trace()
        #if not isinstance(srid, six.integer_types):
            #raise TypeError('An integer SRID must be provided.')
        field_name = kwargs.get('field_name', None)
        tmp, geo_field = self._spatial_setup('transform', field_name=field_name)

        # Getting the selection SQL for the given geographic field.
        field_col = self._geocol_select(geo_field, field_name)

        # Why cascading substitutions? Because spatial backends like
        # Oracle and MySQL already require a function call to convert to text, thus
        # when there's also a transformation we need to cascade the substitutions.
        # For example, 'SDO_UTIL.TO_WKTGEOMETRY(SDO_CS.TRANSFORM( ... )'
        geo_col = self.query.custom_select.get(geo_field, field_col)

        # Setting the key for the field's column with the custom SELECT SQL to
        # override the geometry column returned from the database.
        custom_sel = '%s(%s, %s)' % (connection.ops.transform, geo_col, srid)
        # TODO: Should we have this as an alias?
        # custom_sel = '(%s(%s, %s)) AS %s' % (SpatialBackend.transform, geo_col, srid, qn(geo_field.name))
        self.query.transformed_srid = srid # So other GeoQuerySet methods
        self.query.custom_select[geo_field] = custom_sel
        return self._clone()

    #def _spatial_attribute(self, att, settings, field_name=None, model_att=None):
    def __spatial_attribute(self, *args, **kwargs):
        import ipdb; ipdb.set_trace()
        extra = super(GeoQuerySet, self)._spatial_attribute(*args, **kwargs)
        return extra



import collections

from rest_framework.utils.encoders import JSONEncoder
from spillway.compat import json

#def coerce_feature(feat):
def as_feature(featurelike):
    try:
        return Feature(**featurelike)
    except TypeError:
        #return Feature(**getattr(featurelike, '__geo_interface__', {}))
        return Feature(**featurelike.__geo_interface__)


class LinkedCRS(dict):
    def __init__(self, srid=4326, iterable=(), **kwargs):
        self['type'] = 'link'
        if isinstance(srid, int):
            properties = {}
            properties['href'] = 'http://spatialreference.org/ref/epsg/%s/proj4/' % srid
            properties['type'] = 'proj4'
            self['properties'] = properties
        else:
            iterable = iterable or srid
        self.update(iterable, **kwargs)


#class AbstractFeature(dict):
    #@property
    #def __geo_interface__(self):
        #return self

    #def __set_attribute__(self, ):
        #if k == 'crs':
            #self['crs'] = LinkedCRS(srid)

    # FIXME: Need to override to return Feature and FeatureClass instead of
    # dict!
    #def copy(self):

    #@classmethod
    #def make(cls, feat):


class NamedCRS(dict):
    def __init__(self, srid=4326, iterable=(), **kwargs):
        self['type'] = 'name'
        if isinstance(srid, int):
            self['properties'] = {'name': 'urn:ogc:def:crs:EPSG::%s' % srid}
        else:
            iterable = iterable or srid
        self.update(iterable, **kwargs)


class Feature(dict):
    """GeoJSON Feature dict."""

    def __init__(self, id=None, geometry=None, properties=None,
                 crs=None, iterable=(), **kwargs):
        super(Feature, self).__init__()
        self['type'] = self.__class__.__name__
        self['id'] = id
        self['geometry'] = geometry or {}
        self['properties'] = properties or {}
        if crs:
            self['crs'] = NamedCRS(crs)
        self.update(iterable, **kwargs)

    @property
    def __geo_interface__(self):
        return self

    def __str__(self):
        geom = self['geometry'] or '{}'
        if isinstance(geom, dict):
            return json.dumps(self, cls=JSONEncoder)
        keys = self.viewkeys() - {'geometry'}
        props = json.dumps({k: self[k] for k in keys}, cls=JSONEncoder)[1:-1]
        feature = '{"geometry": %s, %s}' % (geom, props)
        return feature


# FIXME: any way to call __str__ on json.dumps?
class FeatureCollection(dict):
    """GeoJSON FeatureCollection dict."""

    def __init__(self, features=None, crs=None, iterable=(), **kwargs):
        super(FeatureCollection, self).__init__()
        self['type'] = self.__class__.__name__
        if crs:
            self['crs'] = NamedCRS(crs)
        if features and not isinstance(features[0], Feature):
            self['features'] = [Feature(**feat) for feat in features]
            #self['features'] = map(as_feature, features)
            # FIXME: Try **feat AND feat.__geo_interface__
            #try:
                #self['features'] = [Feature(**feat) for feat in features]
            #except TypeError:
                #self['features'] = [Feature(**getattr(feat, '__geo_interface__', {})) for feat in features]
        else:
            self['features'] = features or []
        self.update(iterable, **kwargs)

    @property
    def __geo_interface__(self):
        return self

    def __str__(self):
        features = ','.join(map(str, self['features']))
        keys = self.viewkeys() - {'features'}
        collection = '%s, "features": [' % json.dumps(
            {k: self[k] for k in keys}, cls=JSONEncoder)[:-1]
        return ''.join([collection, features, ']}'])

    #def _coerce_features(self, features):
    #def _coerce_feature(self, features):



class QueryFormMixin(object):
    """Mixin to provide form based handling of GET or POST requests."""
    query_form_class = None

    def get_query_form(self):
        """Returns a bound form instance."""
        return self.query_form_class(
            self.request.QUERY_PARAMS or self.request.DATA,
            self.request.FILES or None)

    def clean_params(self):
        """Returns a validated form dict or an empty dict."""
        form = self.get_query_form()
        return form.cleaned_data if form.is_valid() else {}



from django.contrib.gis import gdal, forms
from django.contrib.gis.db.models.sql.query import ALL_TERMS
from greenwich.srs import transform_tile

from spillway import styles
from spillway.forms import fields


class SpatialQueryForm(forms.Form):
    """A Form for spatial lookup queries such as intersects, overlaps, etc.

    Includes 'bbox' as an alias for 'bboverlaps'.
    """
    bbox = fields.BoundingBoxField(required=False)

    def __init__(self, *args, **kwargs):
        super(SpatialQueryForm, self).__init__(*args, **kwargs)
        for lookup in self.data:
            if lookup in ALL_TERMS:
                self.fields[lookup] = forms.GeometryField(required=False)
                break

    #def full_clean(self):
        #try:
            #return super(SpatialQueryForm, self).full_clean()
        #except:
            #pass

    def clean(self):
        cleaned_data = super(SpatialQueryForm, self).clean()
        spatial_lookup = set(cleaned_data.keys()) - {'bbox'}
        bbox = cleaned_data.pop('bbox', None)
        # Look for "bbox" which is just an alias to "bboverlaps".
        if bbox:
            cleaned_data.pop(spatial_lookup, None)
            cleaned_data['bboverlaps'] = bbox
        return cleaned_data


class GeometryQueryForm(forms.Form):
#class GeoQueryForm(forms.Form): or GeoQuerySetForm?
    """A form providing GeoQuerySet method arguments."""
    # Tolerance value for geometry simplification
    simplify = forms.FloatField(required=False)
    srs = fields.SpatialReferenceField(required=False)


#GeometryRelationships
class GeometryRelationships(forms.Form):
    centroid = forms.BooleanField(required=False)
    envelope = forms.BooleanField(required=False)
    point_on_surface = forms.BooleanField(required=False)
    #op = forms.ChoiceField()

class GeometryOperations(forms.Form):
    #difference
    intersection = forms.GeometryField(required=False)
    #sym_difference
    union = forms.GeometryField(required=False)


class RasterQueryForm(forms.Form):
    """Validates format options for raster data."""
    band = forms.IntegerField(required=False, initial=1)
    bbox = fields.BoundingBoxField(required=False)
    g = fields.OGRGeometryField(required=False)
    upload = fields.GeometryFileField(required=False)

    def clean(self):
        """Return cleaned fields as a dict, determine which geom takes
        precedence.
        """
        cleaned = super(RasterQueryForm, self).clean()
        cleaned['g'] = (cleaned.pop('upload') or cleaned.pop('g') or
                        cleaned.pop('bbox'))
        return cleaned


#class TileSettings(forms.Form):
class MapTile(forms.Form):
    """Validates requested map tiling parameters."""
    #mapfile = forms.FileField(required=False, initial='map.xml')
    bbox = fields.OGRGeometryField(required=False)
    x = forms.IntegerField()
    y = forms.IntegerField()
    z = forms.IntegerField()
    band = forms.IntegerField(required=False, initial=1)
    size = forms.IntegerField(required=False, initial=256)
    style = forms.ChoiceField(
        choices=[(k, k.lower()) for k in list(styles.colors)],
        required=False)
    #opacity = forms.FloatField(min_value=0, max_value=1, required=False)
    #layer =
    #srs = fields.SpatialReferenceField(required=False)

    def clean(self):
        cleaned = super(MapTile, self).clean()
        x, y, z = map(cleaned.get, ('x', 'y', 'z'))
        # Create bbox from NW and SE tile corners.
        extent = transform_tile(x, y, z) + transform_tile(x + 1, y + 1, z)
        geom = gdal.OGRGeometry.from_bbox(extent)
        geom.srid = self.fields['bbox'].default_srid
        cleaned['bbox'] = geom
        return cleaned



from django.contrib.gis.db import models
from rest_framework.filters import BaseFilterBackend

from spillway import forms


class GeoQuerySetFilter(BaseFilterBackend):
    """A Filter for calling GeoQuerySet methods."""
    precision = 4

    def filter_queryset(self, request, queryset, view):
        #form = forms.GeometryQueryForm(request.QUERY_PARAMS)
        #params = form.cleaned_data if form.is_valid() else {}
        params = view.clean_params()
        tolerance, srs = map(params.get, ('simplify', 'srs'))
        srid = getattr(srs, 'srid', None)
        #kwargs = {}
        #kwargs = {'precision': self.precision,
                  #'format': request.accepted_renderer.format)
        #kwargs = {'precision': self.precision, 'format': 'geojson'}
        kwargs = {'precision': self.precision}
        # TODO: Use format='geojson' as a default?
        if not view.wants_default_renderer():
            #kwargs.update(precision=self.precision,
                          #format=request.accepted_renderer.format)
            kwargs.update(format=request.accepted_renderer.format)
        return queryset.simplify(tolerance, srid, **kwargs)


class SpatialLookupFilter(BaseFilterBackend):
    """A Filter providing backend supported spatial lookups like intersects,
    overlaps, etc.
    """

    def filter_queryset(self, request, queryset, view):
        form = forms.SpatialQueryForm(request.QUERY_PARAMS)
        params = form.cleaned_data if form.is_valid() else {}
        #return queryset.filter_geometry(**params)
        qs = queryset.filter_geometry(**params)
        #import ipdb; ipdb.set_trace()
        return qs


class RasterFilter(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        params = view.context.get('params', {})
        geom = params.get('g')
        #request = self.context.get('request')
        format = request.accepted_renderer.format
        return queryset.clip(geom, format)



from .renderers import (GeoJSONRenderer, TemplateRenderer, KMLRenderer,
                        KMZRenderer, SVGRenderer)
from .gdal import (GeoTIFFRenderer, GeoTIFFZipRenderer, HFARenderer,
                   HFAZipRenderer)
from .mapnik import MapnikRenderer, MapnikJPEGRenderer
